#[derive(Clone, Copy)]
struct Header {
    index:isize,
    start:usize,
    length:usize,
}

#[derive(Clone)]
pub struct Sparse<T: Clone> {
    headers:Vec<Header>,
    data:Vec<T>,
}
impl<T: Clone> Sparse<T> {
    pub const fn new() -> Self
    {
        Self {
            headers:Vec::new(),
            data:Vec::new(),
        }
    }

    pub fn set(&mut self, index:isize, data:T)
    // Insert an element at an index.
    {
        match self.find_header(index) {
            Some(hi) => {
                let rel_index = (index - self.headers[hi].index) as usize;
                if rel_index < self.headers[hi].length {

                    // Write over existing element.
                    self.data[self.headers[hi].start + rel_index] = data;
                } else if rel_index == self.headers[hi].length {

                    // Append data to header.
                    self.headers[hi].length += 1;
                    self.data.insert(self.headers[hi].start + rel_index, data);

                    // Check for merge with next header.
                    if hi + 1 < self.headers.len() && self.headers[hi + 1].index == self.headers[hi].index + self.headers[hi].length as isize {
                        self.headers[hi].length += self.headers[hi + 1].length;
                        self.headers.remove(hi + 1);
                    }

                    // Update following headers with new data start position.
                    for hin in (hi + 1)..self.headers.len() {
                        self.headers[hin].start += 1;
                    }
                } else if hi + 1 < self.headers.len() && index == self.headers[hi + 1].index - 1 {

                    // Merge into next header.
                    self.headers[hi + 1].index -= 1;
                    self.headers[hi + 1].length += 1;

                    self.data.insert(self.headers[hi + 1].start, data);

                    // Update following headers with new data start position.
                    for hin in (hi + 2)..self.headers.len() {
                        self.headers[hin].start += 1;
                    }
                } else {

                    // Insert header after current header.
                    let header = Header {
                        index:index,
                        start:self.headers[hi].start + self.headers[hi].length,
                        length:1,
                    };

                    self.data.insert(header.start, data);


                    if hi + 1 >= self.headers.len() {
                        self.headers.push(header);
                    } else {
                        // Increase start index of subsequent headers.
                        for i in (hi + 1)..self.headers.len() {
                            self.headers[i].start += 1;
                        }

                        self.headers.insert(hi + 1, header);
                    }
                }
            }
            None => {
                if self.headers.len() > 0 && index + 1 == self.headers[0].index {
                    // Merge header into first.
                    self.headers[0].index -= 1;
                    self.headers[0].length += 1;
                } else {
                    // Insert header to start of list.
                    for head in &mut self.headers {
                        head.start += 1;
                    }
                    self.headers.insert(0, Header {
                        index:index,
                        start:0,
                        length:1,
                    });
                }
                if self.data.is_empty() {
                    self.data.push(data);
                } else {
                    self.data.insert(0, data);
                }
            }
        }
    }

    pub fn unset(&mut self, index:isize)
    {
        match self.find_header(index) {
            Some(hi) => {
                let rel_index = (index - self.headers[hi].index) as usize;
                if rel_index < self.headers[hi].length {
                    if rel_index == 0 {
                        // Remove from start.
                        self.headers[hi].index += 1;
                        self.headers[hi].length -= 1;
                    } else if rel_index == self.headers[hi].length - 1 {
                        // Remove from end.
                        self.headers[hi].length -= 1;
                    } else {
                        // Split header in two.
                        let header = Header {
                            index:index + 1,
                            start:self.headers[hi].start + rel_index + 1,
                            length:self.headers[hi].length - (rel_index + 1),
                        };

                        self.headers[hi].length -= header.length + 1;

                        self.headers.insert(hi + 1, header);
                    }

                    self.data.remove(self.headers[hi].start + rel_index);

                    // Update following headers with new data start position.
                    for hin in (hi + 1)..self.headers.len() {
                        self.headers[hin].start -= 1;
                    }

                    // Remove header if now empty.
                    if self.headers[hi].length == 0 {
                        self.headers.remove(hi);
                    }
                }
            }
            None => { /* No work to do. */ }
        }
    }

    pub fn get(&self, index:isize) -> Option<&T>
    // Get the element at an index.
    {
        return match self.find_header(index) {
            Some(hi) => {
                if index < self.headers[hi].index + self.headers[hi].length as isize {
                    Some(&self.data[self.headers[hi].start + (index - self.headers[hi].index) as usize])
                } else {
                    None
                }
            }
            None => None
        }
    }

    pub fn get_mut(&mut self, index:isize) -> Option<&mut T>
    // Get the element at an index.
    {
        return match self.find_header(index) {
            Some(hi) => {
                if index < self.headers[hi].index + self.headers[hi].length as isize {
                    Some(&mut self.data[self.headers[hi].start + (index - self.headers[hi].index) as usize])
                } else {
                    None
                }
            }
            None => None
        }
    }

    pub fn data(&self) -> &Vec<T>
    {
        &self.data
    }

    pub fn data_mut(&mut self) -> &mut Vec<T>
    {
        &mut self.data
    }

    pub fn keys(&self) -> Vec<isize>
    {
        let mut list = Vec::<isize>::with_capacity(self.data.len());
        for head in &self.headers {
            for i in 0..head.length {
                list.push(head.index + i as isize);
            }
        }
        return list;
    }

    pub fn pairs(&self) -> Vec<(isize, &T)>
    {
        let mut list = Vec::<(isize,&T)>::with_capacity(self.data.len());
        for head in &self.headers {
            for i in 0..head.length {
                list.push((head.index + i as isize, &self.data[head.start + i]));
            }
        }
        return list;
    }

    pub fn len(&self) -> usize
    {
        self.data.len()
    }

    pub fn print(&self) {
        for head in &self.headers {
            println!("[{:x} -> {}:{}]", head.index, head.start, head.length);
        }
    }

    fn find_header(&self, index:isize) -> Option<usize>
    // Get the index of the header preceeding the key index.
    {
        // Determine if element is before, inside, or after range
        let mut current = self.headers.len() / 2;

        if self.headers.len() > 0 {
            match (index >= self.headers[0].index) as u32
                + (index >= self.headers[self.headers.len() - 1].index as isize) as u32
            {
                1 => {
                    let mut min = 0;
                    let mut max = self.headers.len();

                    while min + 1 != max {
                        if index >= self.headers[current].index && (current + 1 >= max || index < self.headers[current + 1].index) {
                            return Some(current);
                        } else {
                            if index < self.headers[current].index {
                                max = current;
                            } else {
                                min = current;
                            }
                            current = ((max - min) / 2) + min;
                        }
                    }

                    Some(min)
                }
                2 => {
                    Some(self.headers.len() - 1)
                }
                _ => {
                    None
                }
            }
        } else {
            None
        }
    }
}
